package scala.events.test

import org.scalatest.FunSuite
import scala.collection.mutable.Stack
import org.junit.runner.RunWith
import org.scalatest.junit.JUnitRunner
import scala.events.ImperativeEvent
import scala.events.behaviour.Signal
import org.scalatest.matchers.ShouldMatchers
import scala.events.scalareact
import scala.events.behaviour.Var

@RunWith(classOf[JUnitRunner])
class SignalSuite extends FunSuite with ShouldMatchers {

  test("EScala events can be accumulated with fold") {
    val e = new ImperativeEvent[Int]
    val accum: Signal[Int] = e.fold(0) { _ + _ }
    e(1)
    e(2)
    accum.getValue should be(3)
    e(3)
    accum.getValue should be(6)
  }

  test("Signals have an EScala changed event") {
    val e = new ImperativeEvent[Int]
    val accum: Signal[Int] = e.fold(0) { _ + _ }

    var mutableString = ""

    accum.changed += { a => mutableString = a.toString }
    e(1)
    mutableString should be("1")
    e(2)
    mutableString should be("3")
    e(3)
    mutableString should be("6")
  }

  test("Scala.react dataflow can be used to create signals") {
    val e = new ImperativeEvent[Int]
    val accum = e.fold(0) { _ + _ }

    val flowSignal = scalareact.Signal.flow("No occurence") { self =>
      self await accum
      self() = "First occurence"
      self awaitNext accum
      self() = "Second occurence"
    }

    flowSignal.getValue should be("No occurence")
    e(1)
    flowSignal.getValue should be("First occurence")
    e(1)
    flowSignal.getValue should be("Second occurence")
    e(1)
    flowSignal.getValue should be("Second occurence")

  }
  
  test("Dependand signals are updated transparently") {
    val e = new ImperativeEvent[Int]
    val accum: Signal[Int] = e.fold(0) { _ + _ }
    val dependant = Signal {accum() * 2}
    e(1)
    e(2)
    e(3)
    accum.getValue should be(6)
    dependant.getValue should be(12)
  }
  
  test("A Signal can depend on Vars"){
    val a = Var(1)
    val b = Var(2)
    val signal = Signal { a() + b() }
    val other = Signal { a() + 1 }
    
    signal.getValue should be (3)
    
    a() = 2
    signal.getValue should be (4)
  }
  
  test("Vars have a changed event which gets triggered on value change"){
    val a = Var(1)
    val nChanges = a.changed.iterate(0)(_ + 1)
    a() = 23 // 1 change
    a() = 42 // 2 changes
    a() = 42 // no change
    
    nChanges.getValue should be (2)
  }
  
  test("Changed events get propageted to dependants"){
    val a = Var(1)
    val dep = Signal { a() + 1 }
    dep.getValue should be (2)
    
    var mutable = 0
    dep.changed += {mutable = _}
    
    mutable should be (0)
    a() = 2
    mutable should be (3)
  }
  
   test("Dependant signals receive change events"){
    val a = Var(0)
    a.changed.iterate(0){1 + _}
    val dep = Signal { a() }
    
    // when removing this line, test fails, for some reason
    Signal { a() }  
    
    var mutable = 0
    dep.changed += {mutable = _}
    
    dep.getValue should be (0)
    a() = 2
    dep.getValue should be (2)
    mutable should be (2)
  }
  
  
  test("All handlers are called once per event occurence"){
    val a = Var(0)
    val dep = Signal { a() + 1 }
    
    var mutable = 0
    a.changed += {_ => mutable += 1}
    a.changed += {_ => mutable += 1}
    dep.changed += {_ => mutable += 1}
    
    // should trigger a changed event 2 times. 
    a() = 42
    a() = 42
    a() = 0
    // each of the 3 handlers should be called once
    mutable should be (6)
  }
  
  test("iterate can be used to count event occurences"){
    val e = new ImperativeEvent[Unit]
    val n = e.iterate(0)(_ + 1)
    n.getValue should be (0)
    e()
    n.getValue should be (1)
  }
  
  
  test("reactive datastructures (SStack)"){
    import scala.events.behaviour.datastructures._;
    
    var sum = 0
    val sstack = SStack(5,10,10,11)
    val top = sstack.top
    
    
    top.changed += { _ match { 
	      case Some(x) => sum += x
	      case None => sum = 0
    	}
     }
   
    // Important note: top.changed obviously does not get called for each pop(), but only when the top value changes    
    sstack.top.getValue should be (Some(11))
    sstack.pop // top changed to 10
    sstack.pop // top does not change
    sstack.pop // top changed to 5
    sum should be (15)
    sstack.top.getValue should be (Some(5))
    sstack.pop
    sum should be (0)
    sstack.top.getValue should be (None)
    
    sstack.push(3)
    sum should be (3)
    sstack.top.getValue should be (Some(3))
  }
  
  
   test("reactive datastructures (SQueue)"){
    import scala.events.behaviour.datastructures._;
    val squeue = SQueue(1, 1, 1)
    
    var nLengthChanges = squeue.length.changed.iterate(0)(_ + 1)
    
    var empty = squeue.isEmpty.getValue
    squeue.isEmpty.changedTo(true) += {_ =>
      empty = true
    }
    
    empty should be (false)
    
    nLengthChanges.getValue should be (0)    
    squeue.length.getValue should be (3)
    squeue.dequeue
    squeue.length.getValue should be (2)
    nLengthChanges.getValue should be (1)
    squeue.dequeue
    squeue.length.getValue should be (1)
    nLengthChanges.getValue should be (2)
    squeue.enqueue(1)
    squeue.length.getValue should be (2)
    nLengthChanges.getValue should be (3)
    
    squeue.dequeue
    squeue.dequeue
    empty should be (true)
    
   }
  
  
 ignore("Scala.react turns get triggered at most 2 * [event invocation + signal instanciation]") {
    val e = new ImperativeEvent[Int]

    val eventCounter = e.iterate(0)(_ + 1)
    val accum = e.fold(0) { _ + _ }
    val dep1 = Signal { accum() + 1 }
    val dep2 = Signal { dep1() * 2 }

    e(1)
    e(2)
    e(3)

    val nEvents = 3
    val nRootSignals = 2
    val maxTurns = 2 * (nEvents + nRootSignals)
    assert(eventCounter.getValue === nEvents)

    val turns = scala.events.scalareact.engine.currentTurn.asInstanceOf[Int]
    turns should be <= (maxTurns)
  }


}